import logging
import signal

import django
from django.core.management.base import BaseCommand, CommandParser
from django.db import close_old_connections

import arrow
import rollbar
from mbq.pubsub import _collector as collector
from mbq.pubsub import models, utils
from mbq.pubsub.settings import project_settings


logger = logging.getLogger(__name__)


class SignalHandler:
    def __init__(self):
        self._interrupted = False

    def handle_signal(self, *args, **kwargs):
        self._interrupted = True

    def should_continue(self):
        return not self._interrupted


class Command(BaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.signal_handler = SignalHandler()
        signal.signal(signal.SIGINT, self.signal_handler.handle_signal)
        signal.signal(signal.SIGTERM, self.signal_handler.handle_signal)
        signal.signal(signal.SIGQUIT, self.signal_handler.handle_signal)

    def add_arguments(self, parser):
        # django fixed a long-standing annoyance in 2.1 so we only
        # need this hack if we're using a version before that.
        # SubParser inspired by:
        # https://stackoverflow.com/a/37414551/305736
        add_parsers_kwargs = {}
        if django.VERSION < (2, 1):
            command = self

            class SubParser(CommandParser):
                def __init__(self, **kwargs):
                    return super().__init__(command, **kwargs)

            add_parsers_kwargs["parser_class"] = SubParser

        queue_choices = sorted(project_settings.QUEUES)

        subparsers = parser.add_subparsers(dest="command", **add_parsers_kwargs)
        # we should be able to pass this to .add_subparsers above but
        # that kwarg is broken in Python 3.6
        subparsers.required = True

        consume_parser = subparsers.add_parser("consume")
        consume_parser.add_argument("-q", "--queue", required=True, choices=queue_choices)

        replay_parser = subparsers.add_parser("replay")
        replay_parser.add_argument("-q", "--queue", required=True, choices=queue_choices)
        replay_parser.add_argument("num_messages", type=int)

    def handle(self, **options):
        getattr(self, f'handle_{options["command"]}')(**options)

    def handle_consume(self, queue=None, **options):
        try:
            consumer = project_settings.CONSUMER_CLASS(
                queue,
                project_settings.MESSAGE_HANDLERS,
                default_handler=project_settings.DEFAULT_HANDLER,
            )
            while self.signal_handler.should_continue():
                consumer.process_queue()
                if project_settings.USE_DATABASE:
                    if utils.is_db_ready():
                        consumer.process_dead_letter_queue()
                        self._send_dlq_metrics(queue)
                        self._delete_old_dlq_messages(queue)
                    else:
                        logger.info(
                            "Database is not ready. Skipping processing the DLQ and collecting metrics."
                        )

            # django only cleans up db connections when handling the "request_finished" signal
            # so we want to make sure it happens regularly here in the long-lived consumer
            close_old_connections()
        except Exception:
            logger.exception("Consume failed")
            rollbar.report_exc_info()

    def handle_replay(self, queue=None, num_messages=None, **options):
        consumer = project_settings.CONSUMER_CLASS(
            queue,
            project_settings.MESSAGE_HANDLERS,
            default_handler=project_settings.DEFAULT_HANDLER,
        )
        consumer.replay_dead_letter_queue(num_messages)

    def _send_dlq_metrics(self, queue):
        collector.gauge(
            "consumer.undeliverable_messages.total",
            models.UndeliverableMessage.objects.filter(queue=queue).count(),
            tags={"queue": queue},
        )

    @utils.debounce(minutes=15)
    def _delete_old_dlq_messages(self, queue):
        days_ago_30 = arrow.utcnow().shift(days=-30)
        models.UndeliverableMessage.objects.filter(
            queue=queue, created_at__lte=days_ago_30.datetime
        ).delete()



from django.apps import AppConfig

from mbq import metrics
from mbq.pubsub.settings import project_settings


class PubSubConfig(AppConfig):
    name = "mbq.pubsub"
    verbose_name = "PubSub"

    def ready(self):
        self.module._collector = metrics.Collector(
            namespace="mbq.pubsub",
            tags={"env": project_settings.ENV.long_name, "service": project_settings.SERVICE},
        )



import json
import urllib.parse

from django.conf import settings
from django.contrib import admin
from django.utils.html import format_html

import boto3
from botocore.exceptions import ClientError
from mbq.pubsub import models, utils


def get_name_from_topic_arn(arn):
    if arn:
        return arn.split(":")[-1]


class SNSTopicListFilter(admin.SimpleListFilter):
    title = "topic"
    parameter_name = "topic_arn"

    def lookups(self, request, model_admin):
        qs = model_admin.get_queryset(request).order_by("topic_arn")
        topic_arns = qs.values_list("topic_arn", flat=True).distinct()
        for arn in topic_arns:
            yield (arn, get_name_from_topic_arn(arn))

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(topic_arn=self.value())
        return queryset


def replay_messages(model_admin, request, queryset):
    sqs = boto3.client("sqs")
    queue_urls = {}

    for message in queryset:
        if message.queue not in queue_urls:
            url = sqs.get_queue_url(QueueName=utils.construct_full_queue_name(message.queue))
            queue_urls[message.queue] = url["QueueUrl"]

        try:
            sqs.send_message(QueueUrl=queue_urls[message.queue], MessageBody=message.payload)
            message.delete()
        except ClientError as e:
            model_admin.message_user(
                request, f"Message {message.id} failed to replay: {e.response}"
            )


replay_messages.short_description = "Replay messages"


class UndeliverableMessageAdmin(admin.ModelAdmin):
    list_display = ("id", "created_at", "message_timestamp", "message_type", "queue")

    list_filter = ("message_type", "queue")

    fields = ("id", "created_at", "message_timestamp", "message_type", "queue", "topic_arn")

    readonly_fields = (
        "id",
        "created_at",
        "message_timestamp",
        "message_type",
        "queue",
        "topic_arn",
    )

    def has_delete_permission(self, *args):
        return True

    def has_add_permission(self, request):
        return False

    def admin_topic(self, message):
        return get_name_from_topic_arn(message.topic_arn)

    admin_topic.short_description = "topic"

    list_display = list_display + ("admin_topic",)
    list_filter = list_filter + (SNSTopicListFilter,)

    def admin_papertrail_link(self, message):
        papertrail_link = settings.PUBSUB.get("PAPERTRAIL_URL")
        if papertrail_link:
            message_id = json.loads(message.payload)["MessageId"]
            url = f"{papertrail_link}/events?q={message_id}"
            return format_html('<a href="{url}" target="_blank">{text}</a>', url=url, text=url)
        else:
            return "To see a papertrai link, add the 'PAPERTRAIL_URL' setting to your ranch config"

    admin_papertrail_link.short_description = "papertrail"

    def admin_rollbar_query(self, message):
        message_id = json.loads(message.payload)["MessageId"]
        query = urllib.parse.quote(f"context:pubsub-message-id#{message_id}")
        service = settings.PUBSUB.get("SERVICE")
        url = f"https://rollbar.com/ManagedByQ/{service}/items/?query={query}"
        return format_html('<a href="{url}" target="_blank">{text}</a>', url=url, text=url)

    admin_rollbar_query.short_description = "rollbar"

    def admin_payload(self, message):
        try:
            payload = json.loads(message.payload)
            data = json.loads(payload["Message"])
            payload["Message"] = data
            payload_str = json.dumps(payload, indent=4)
            return format_html("<br/><pre>{}</pre>", payload_str)
        except Exception:
            return format_html("<br/><pre>{}</pre>", message.payload)

    admin_payload.short_description = "payload"

    fields = fields + ("admin_payload", "admin_papertrail_link", "admin_rollbar_query")
    readonly_fields = readonly_fields + (
        "admin_payload",
        "admin_papertrail_link",
        "admin_rollbar_query",
    )

    actions = [replay_messages]


admin.site.register(models.UndeliverableMessage, UndeliverableMessageAdmin)



import functools
import time

from django.db import connections
from django.db.migrations.executor import MigrationExecutor

from .settings import project_settings


_DB_READY = {}


def construct_full_queue_name(queue_name):
    return f"mbq-{project_settings.SERVICE}-{queue_name}-{project_settings.ENV.short_name}"


def debounce(seconds=None, minutes=None, hours=None):
    def wrapper(func):
        func.seconds_between_runs = 0
        func.last_run = time.time()

        if seconds:
            func.seconds_between_runs += seconds
        if minutes:
            func.seconds_between_runs += minutes * 60
        if hours:
            func.seconds_between_runs += hours * 60 * 60

        @functools.wraps(func)
        def wrapped_func(*args, **kwargs):
            if func.last_run + func.seconds_between_runs < time.time():
                func(*args, **kwargs)
                func.last_run = time.time()

        return wrapped_func

    return wrapper


def is_db_ready(database="default"):
    """Determine whether the migrations for pubsub are up to date. Can
    be used to defer hitting the database until everything is ready.

    Implementation is inspired by the `./manage.py migrate --plan` command:

       https://github.com/django/django/blob/master/django/core/management/commands/migrate.py#L140-L150
    """
    global _DB_READY

    if not _DB_READY.get(database, False):
        executor = MigrationExecutor(connections[database])
        # find the pubsub migrations
        pubsub_migrations = [
            node for node in executor.loader.graph.leaf_nodes() if node[0] == "pubsub"
        ]
        # build a plan to run all the migrations
        plan = executor.migration_plan(pubsub_migrations)
        # if there's no plan then we're fully up to date
        _DB_READY[database] = not bool(plan)

    return _DB_READY[database]



import json
import logging
import typing

import arrow
import boto3
import rollbar

from . import _collector as collector
from . import constants, exceptions, models, utils


logger = logging.getLogger(__name__)

NOT_PROVIDED = object()


class Consumer:
    def __init__(
        self,
        queue_name: str,
        handlers: dict,
        default_handler: typing.Optional[typing.Callable[[str], None]] = None,
    ):
        self._queue_name = queue_name
        self._queue_full_name = utils.construct_full_queue_name(queue_name)
        self._handlers = handlers
        self._default_handler = default_handler

    @property
    def queue(self):
        if not hasattr(self, "_queue"):
            sqs = boto3.resource("sqs")
            self._queue = sqs.get_queue_by_name(QueueName=self._queue_full_name)
        return self._queue

    @property
    def dead_letter_queue(self):
        """
        Find the dead letter queue from the primary queue's redrive policy.

        https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-dead-letter-queues.html
        """

        if not hasattr(self, "_dead_letter_queue"):
            try:
                redrive_policy = json.loads(self.queue.attributes["RedrivePolicy"])
                dlq_name = redrive_policy["deadLetterTargetArn"].split(":")[-1]
            except Exception as e:
                raise exceptions.ConsumerException(
                    f"No dead letter queue configured for {self.queue}"
                ) from e

            sqs = boto3.resource("sqs")
            self._dead_letter_queue = sqs.get_queue_by_name(QueueName=dlq_name)
        return self._dead_letter_queue

    def process_queue(self):
        collector.increment("consumer.attempt_read", tags={"queue": self._queue_name})
        messages = self.queue.receive_messages(WaitTimeSeconds=5, MaxNumberOfMessages=10)
        if len(messages) > 0:
            logger.info(f"Received {len(messages)} messages")

        for message in messages:
            try:
                body = json.loads(message.body)
                data = json.loads(body["Message"])

                message_type = data["message_type"]
                payload = data["payload"]
                handler = self._handlers.get(message_type)
            except Exception:
                if self._default_handler:
                    message_type = constants.DEFAULT_MESSAGE_TYPE
                    payload = message.body
                    handler = self._default_handler
                else:
                    rollbar.report_exc_info()
                    continue
            try:
                if handler:
                    logger.info(f"Processing {message_type} message")
                    handler(payload)
                    result = "succeeded"
                else:
                    result = "skipped"
                    logger.info(f"Received unregistered message_type: {message_type}")
            except Exception:
                result = "failed"
                messageId = body["MessageId"]
                uuid = rollbar.report_exc_info(
                    payload_data={"context": f"pubsub-message-id#{messageId}"}
                )  # is None on local container test
                url = f"https://rollbar.com/item/uuid/?uuid={uuid}"
                logger.exception(
                    f"An error occurred while processing the "
                    f"message. MessageId: {messageId} Rollbar URL: {url}"
                )
            else:
                message.delete()

            collector.increment(
                "consumer.processed",
                tags={"result": result, "message_type": message_type, "queue": self._queue_name},
            )

    def process_dead_letter_queue(self):
        messages = self.dead_letter_queue.receive_messages(
            WaitTimeSeconds=0, MaxNumberOfMessages=10
        )
        if len(messages) > 0:
            logger.info(f"Received {len(messages)} messages on the dead letter queue")

        for message in messages:
            try:
                body = json.loads(message.body)
                data = json.loads(body["Message"])
                message_type = data["message_type"]
            except Exception as e:
                if self._default_handler:
                    message_type = constants.DEFAULT_MESSAGE_TYPE
                else:
                    raise e

            models.UndeliverableMessage.objects.create(
                message_type=message_type,
                message_timestamp=arrow.get(body.get("Timestamp")).datetime,
                payload=message.body,
                queue=self._queue_name,
                topic_arn=body.get("TopicArn"),
            )
            message.delete()

    def replay_dead_letter_queue(self, max_messages: int):
        if max_messages < 1:
            raise exceptions.ConsumerException("max_messages must be greater than 0")

        logger.info(
            f"Replaying at most {max_messages} messages "
            f"from {self.dead_letter_queue} to {self.queue}"
        )

        messages_processed = 0
        while True:
            messages = self.dead_letter_queue.receive_messages(
                WaitTimeSeconds=20, MaxNumberOfMessages=10
            )

            for message in messages:
                self.queue.send_message(MessageBody=message.body)
                message.delete()
                messages_processed += 1

                if messages_processed >= max_messages:
                    return



import json

from django.core.management.base import BaseCommand

import boto3
from mbq import env


class Command(BaseCommand):
    def handle(self, *args, **options):
        sns = boto3.client("sns")
        sns.publish(
            TargetArn=env.get("SNS_ARN"),
            MessageStructure="json",
            Message=json.dumps(
                {
                    "default": json.dumps(
                        {"message_type": "pubsub.update_raise_exception", "payload": {"foo": "bar"}}
                    )
                }
            ),
        )



import uuid

from django.db import models


class UndeliverableMessage(models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)

    created_at = models.DateTimeField(auto_now_add=True)

    message_type = models.CharField(max_length=256)
    message_timestamp = models.DateTimeField()
    payload = models.TextField()
    queue = models.CharField(max_length=80)
    topic_arn = models.CharField(max_length=2048, null=True)



from .__version__ import (  # noqa
    __author__,
    __author_email__,
    __description__,
    __license__,
    __title__,
    __url__,
    __version__,
)


default_app_config = "mbq.pubsub.apps.PubSubConfig"
